<?php

declare(strict_types=1);

namespace OliverKlee\Oelib\Testing;

use Doctrine\DBAL\Driver\ResultStatement;
use OliverKlee\Oelib\Authentication\FrontEndLoginManager;
use OliverKlee\Oelib\DataStructures\Collection;
use OliverKlee\Oelib\FrontEnd\UserWithoutCookies;
use OliverKlee\Oelib\Mapper\FrontEndUserMapper;
use OliverKlee\Oelib\Mapper\MapperRegistry;
use OliverKlee\Oelib\Model\FrontEndUserGroup;
use Psr\Log\NullLogger;
use TYPO3\CMS\Core\Configuration\SiteConfiguration;
use TYPO3\CMS\Core\Context\Context;
use TYPO3\CMS\Core\Context\UserAspect;
use TYPO3\CMS\Core\Core\Environment;
use TYPO3\CMS\Core\Database\Connection;
use TYPO3\CMS\Core\Database\ConnectionPool;
use TYPO3\CMS\Core\Database\Query\QueryBuilder;
use TYPO3\CMS\Core\Exception\Page\PageNotFoundException;
use TYPO3\CMS\Core\Http\ServerRequest;
use TYPO3\CMS\Core\Information\Typo3Version;
use TYPO3\CMS\Core\Localization\Locales;
use TYPO3\CMS\Core\Routing\PageArguments;
use TYPO3\CMS\Core\Site\Entity\Site;
use TYPO3\CMS\Core\TypoScript\TemplateService;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\RootlineUtility;
use TYPO3\CMS\Frontend\Authentication\FrontendUserAuthentication;
use TYPO3\CMS\Frontend\ContentObject\ContentObjectRenderer;
use TYPO3\CMS\Frontend\Controller\TypoScriptFrontendController;

/**
 * This class provides various functions to handle dummy records in unit tests.
 */
final class TestingFramework
{
    /**
     * @var positive-int
     */
    private const AUTO_INCREMENT_THRESHOLD_WITHOUT_ROOTLINE_CACHE = 100;

    /**
     * all system table names to which this instance of the testing framework
     * has access
     *
     * @var list<non-empty-string>
     */
    private const ALLOWED_SYSTEM_TABLES = [
        // @deprecated #1537 will be removed in oelib 6.0
        'be_users',
        'fe_groups',
        'fe_users',
        'pages',
        'sys_template',
        'tt_content',
        'be_groups',
        'sys_file',
        'sys_file_collection',
        'sys_file_reference',
        'sys_category',
        'sys_category_record_mm',
    ];

    /**
     * @var non-empty-string
     */
    private const FAKE_FRONTEND_DOMAIN_NAME = 'typo3-test.dev';

    /**
     * @var non-empty-string
     */
    private const SITE_IDENTIFIER = 'testing-framework';

    /**
     * cache for the results of hasTableColumn with the column names as keys and
     * the SHOW COLUMNS field information (in an array) as values
     *
     * @var array<string, array<string, array<string, string>>>
     */
    private static $tableColumnCache = [];

    /**
     * @var array<non-empty-string, array<string, string|int|null>> cache for the results of existsTable with the
     *      table names as keys and the table SHOW STATUS information (in an array) as values
     */
    private static $tableNameCache = [];

    /**
     * @var bool
     */
    private $databaseInitialized = false;

    /**
     * prefix of the extension for which this instance of the testing framework
     * was instantiated (e.g. "tx_seminars")
     *
     * @var non-empty-string
     */
    private $tablePrefix;

    /**
     * all own DB table names to which this instance of the testing framework has access
     *
     * @var list<non-empty-string>
     */
    private $ownAllowedTables = [];

    /**
     * all "dirty" non-system tables (i.e., all tables that were used for testing
     * and need to be cleaned up)
     *
     * @var array<string, non-empty-string>
     *
     * @deprecated #1532 will be removed in oelib 6.0
     */
    private $dirtyTables = [];

    /**
     * all "dirty" system tables (i.e. all tables that were used for testing and
     * need to be cleaned up)
     *
     * @var array<string, non-empty-string>
     *
     * @deprecated #1532 will be removed in oelib 6.0
     */
    private $dirtySystemTables = [];

    /**
     * sorting values of all relation tables
     *
     * @var array<non-empty-string, array<positive-int, int<0, max>>>
     */
    private $relationSorting = [];

    /**
     * The number of unusable UIDs after the maximum UID in a table before the auto increment value will be reset by
     * resetAutoIncrementLazily.
     *
     * This value needs to be high enough so that no two page UIDs will be the same within on request as the local
     * root-line cache of TYPO3 CMS otherwise might create false cache hits, causing failures for unit tests relying on
     * the root line.
     *
     * @see https://bugs.oliverklee.com/show_bug.cgi?id=5011
     * @deprecated will be removed in oelib 6.0
     *
     * @var int
     *
     * @deprecated #1529 will be removed in oelib 6.0
     */
    private $resetAutoIncrementThreshold = 0;

    /**
     * whether a fake front end has been created
     *
     * @var bool
     */
    private $hasFakeFrontEnd = false;

    /**
     * hook objects for this class
     *
     * @var list<object>
     */
    private static $hooks = [];

    /**
     * whether the hooks in self::hooks have been retrieved
     *
     * @var bool
     */
    private static $hooksHaveBeenRetrieved = false;

    /**
     * @var array<string, string|bool|null>|null
     */
    private $serverVariablesBackup;

    /**
     * This testing framework can be instantiated for one extension at a time.
     * Example: In your testcase, you'll have something similar to this line of code:
     *
     * `$this->subject = new TestingFramework('tx_seminars');`
     *
     * The parameter you provide is the prefix of the table names of that particular
     * extension. Like this, we ensure that the testing framework creates and
     * deletes records only on table with this prefix.
     *
     * If you need dummy records on tables of multiple extensions, you will have to
     * instantiate the testing framework multiple times (once per extension).
     *
     * Instantiating this class sets all core caches in order to avoid errors about not registered caches.
     *
     * @param non-empty-string $tablePrefix table name prefix of the extension
     *        for this instance of the testing framework
     */
    public function __construct(string $tablePrefix)
    {
        $this->tablePrefix = $tablePrefix;

        (new CacheNullifyer())->setAllCoreCaches();
    }

    private function initializeDatabase(): void
    {
        if ($this->databaseInitialized) {
            return;
        }

        $this->createListOfOwnAllowedTables();
        $this->determineAndSetAutoIncrementThreshold();

        $this->databaseInitialized = true;
    }

    /**
     * Determines a good value for the auto increment threshold and sets it.
     *
     * @deprecated #1529 will be removed in oelib 6.0
     */
    private function determineAndSetAutoIncrementThreshold(): void
    {
        $this->setResetAutoIncrementThreshold(self::AUTO_INCREMENT_THRESHOLD_WITHOUT_ROOTLINE_CACHE);
    }

    /**
     * Creates a new dummy record for unit tests.
     *
     * If no record data for the new array is given, an empty record will be
     * created. It will only contain a valid UID and the "is_dummy_record" flag
     * will be set to 1.
     *
     * Should there be any problem creating the record (wrong table name or a
     * problem with the database), 0 instead of a valid UID will be returned.
     *
     * @param non-empty-string $table the name of the table on which the record should be created
     * @param array<string, string|int|bool> $recordData data to save in the new record, may be empty,
     *        but must not contain the key "uid"
     *
     * @return positive-int the UID of the new record
     *
     * @throws \InvalidArgumentException
     */
    public function createRecord(string $table, array $recordData = []): int
    {
        $this->initializeDatabase();
        if (!$this->isNoneSystemTableNameAllowed($table)) {
            $allowedTables = \implode(',', $this->ownAllowedTables);
            $errorMessage = \sprintf('The table "%1$s" is not allowed. Allowed tables: %2$s', $table, $allowedTables);
            throw new \InvalidArgumentException($errorMessage, 1331489666);
        }
        if (isset($recordData['uid'])) {
            throw new \InvalidArgumentException('The column "uid" must not be set in $recordData.', 1331489678);
        }

        return $this->createRecordWithoutTableNameChecks($table, $recordData);
    }

    /**
     * Creates a new dummy record for unit tests without checks for the table name.
     *
     * If no record data for the new array is given, an empty record will be created.
     * It will only contain a valid UID and the "is_dummy_record" flag will be set to 1.
     *
     * Should there be any problem creating the record (wrong table name or a
     * problem with the database), 0 instead of a valid UID will be returned.
     *
     * @param non-empty-string $table the name of the table on which the record should be created
     * @param array<string, string|int|bool> $rawData data to save, may be empty, but must not contain the key "uid"
     *
     * @return positive-int the UID of the new record
     */
    private function createRecordWithoutTableNameChecks(string $table, array $rawData): int
    {
        $this->initializeDatabase();
        $dataToInsert = $this->normalizeDatabaseRow($rawData);
        // @deprecated #1532 will be removed in oelib 6.0
        $dummyColumnName = $this->getDummyColumnName($table);
        $dataToInsert[$dummyColumnName] = 1;

        $connection = $this->getConnectionForTable($table);
        $connection->insert($table, $dataToInsert);

        // @deprecated #1532 will be removed in oelib 6.0
        $this->markTableAsDirty($table);

        $uid = (int)$connection->lastInsertId($table);
        \assert($uid > 0);

        return $uid;
    }

    /**
     * Normalizes the types in the given data so that the data con be inserted into a DB.
     *
     * @param array<string, string|int|bool|float> $rawData
     *
     * @return array<string, string|int|float>
     */
    private function normalizeDatabaseRow(array $rawData): array
    {
        $dataToInsert = [];
        foreach ($rawData as $key => $value) {
            $dataToInsert[$key] = \is_bool($value) ? (int)$value : $value;
        }

        return $dataToInsert;
    }

    /**
     * @param non-empty-string $tableName
     *
     * @return Connection
     */
    private function getConnectionForTable(string $tableName): Connection
    {
        return $this->getConnectionPool()->getConnectionForTable($tableName);
    }

    /**
     * @param non-empty-string $tableName
     *
     * @return QueryBuilder
     */
    private function getQueryBuilderForTable(string $tableName): QueryBuilder
    {
        return $this->getConnectionPool()->getQueryBuilderForTable($tableName);
    }

    /**
     * @return ConnectionPool
     */
    private function getConnectionPool(): ConnectionPool
    {
        return GeneralUtility::makeInstance(ConnectionPool::class);
    }

    /**
     * Creates a front-end page on the given page, and provides it with the page UID as slug.
     *
     * @param array<string, string|int>|null $data
     *
     * @return positive-int the UID of the new page
     */
    public function createFrontEndPage(int $parentPageUid = 0, ?array $data = null): int
    {
        $data = $data ?? [];
        $hasSlug = \array_key_exists('slug', $data);
        $uid = $this->createGeneralPageRecord(1, $parentPageUid, $data);
        if (!$hasSlug) {
            $this->changeRecord('pages', $uid, ['slug' => '/' . $uid]);
        }

        return $uid;
    }

    /**
     * Creates a system folder on the page with the UID given by the first
     * parameter $parentId.
     *
     * @param int $parentId
     *        UID of the page on which the system folder should be created
     *
     * @return positive-int the UID of the new system folder
     */
    public function createSystemFolder(int $parentId = 0): int
    {
        return $this->createGeneralPageRecord(254, $parentId, []);
    }

    /**
     * Creates a page record with the document type given by the first parameter
     * $documentType.
     *
     * The record will be created on the page with the UID given by the second
     * parameter $parentId.
     *
     * @param positive-int $documentType document type of the record to create
     * @param int $parentId UID of the page on which the record should be created
     * @param array<string, string|int|bool> $recordData data to save in the record, may be empty,
     *        but must not contain the keys "uid", "pid" or "doktype"
     *
     * @return positive-int the UID of the new record
     *
     * @throws \InvalidArgumentException
     */
    private function createGeneralPageRecord(int $documentType, int $parentId, array $recordData): int
    {
        if (isset($recordData['uid'])) {
            throw new \InvalidArgumentException('The column "uid" must not be set in $recordData.', 1331489697);
        }
        if (isset($recordData['pid'])) {
            throw new \InvalidArgumentException('The column "pid" must not be set in $recordData.', 1331489703);
        }
        if (isset($recordData['doktype'])) {
            throw new \InvalidArgumentException('The column "doktype" must not be set in $recordData.', 1331489708);
        }

        $completeRecordData = $recordData;
        $completeRecordData['pid'] = $parentId;
        $completeRecordData['doktype'] = $documentType;

        return $this->createRecordWithoutTableNameChecks('pages', $completeRecordData);
    }

    /**
     * Creates a template on the page with the UID given by the first parameter $pageId.
     *
     * @param positive-int $pageId UID of the page on which the template should be created
     * @param array<string, string|int|bool> $recordData data to save, may be empty,
     *        but must not contain the keys "uid" or "pid"
     *
     * @return positive-int the UID of the new template
     *
     * @throws \InvalidArgumentException
     */
    public function createTemplate(int $pageId, array $recordData = []): int
    {
        // @phpstan-ignore-next-line We're testing for a contract violation here.
        if ($pageId <= 0) {
            throw new \InvalidArgumentException('$pageId must be > 0.', 1331489774);
        }
        if (isset($recordData['uid'])) {
            throw new \InvalidArgumentException('The column "uid" must not be set in $recordData.', 1331489769);
        }
        if (isset($recordData['pid'])) {
            throw new \InvalidArgumentException('The column "pid" must not be set in $recordData.', 1331489764);
        }

        $completeRecordData = $recordData;
        $completeRecordData['pid'] = $pageId;

        return $this->createRecordWithoutTableNameChecks('sys_template', $completeRecordData);
    }

    /**
     * Creates a FE user group.
     *
     * @param array<string, string|int|bool> $recordData data to save, may be empty, but must not contain the key "uid"
     *
     * @return positive-int the UID of the new user group
     *
     * @throws \InvalidArgumentException
     */
    public function createFrontEndUserGroup(array $recordData = []): int
    {
        if (isset($recordData['uid'])) {
            throw new \InvalidArgumentException('The column "uid" must not be set in $recordData.', 1331489807);
        }

        return $this->createRecordWithoutTableNameChecks('fe_groups', $recordData);
    }

    /**
     * Creates a FE user record.
     *
     * @param string|int $frontEndUserGroups
     *        comma-separated list of UIDs of the user groups to which the new user belongs, each must be > 0,
     *        may contain spaces, if empty a new FE user group will be created
     * @param array<string, string|int|bool> $recordData data to save, may be empty,
     *        but must not contain the keys "uid" or "usergroup"
     *
     * @return positive-int the UID of the new FE user
     *
     * @throws \InvalidArgumentException
     */
    public function createFrontEndUser($frontEndUserGroups = '', array $recordData = []): int
    {
        $frontEndUserGroupsWithoutSpaces = str_replace(' ', '', (string)$frontEndUserGroups);

        if ($frontEndUserGroupsWithoutSpaces === '') {
            $frontEndUserGroupsWithoutSpaces = (string)$this->createFrontEndUserGroup();
        }
        $groupsCheckResult = \preg_match('/^(?:[1-9]+\\d*,?)+$/', $frontEndUserGroupsWithoutSpaces);
        if (!\is_int($groupsCheckResult) || $groupsCheckResult === 0) {
            throw new \InvalidArgumentException(
                '$frontEndUserGroups must contain a comma-separated list of UIDs. Each UID must be > 0.',
                1331489824
            );
        }
        if (isset($recordData['uid'])) {
            throw new \InvalidArgumentException('The column "uid" must not be set in $recordData.', 1331489842);
        }
        if (isset($recordData['usergroup'])) {
            throw new \InvalidArgumentException('The column "usergroup" must not be set in $recordData.', 1331489846);
        }

        $completeRecordData = $recordData;
        $completeRecordData['usergroup'] = $frontEndUserGroupsWithoutSpaces;

        return $this->createRecordWithoutTableNameChecks('fe_users', $completeRecordData);
    }

    /**
     * Creates and logs in an FE user.
     *
     * @param string|int $frontEndUserGroups comma-separated list of UIDs of the user groups to which the user belongs,
     *        each must be > 0, may contain spaces; if empty a new front-end user group is created
     * @param array<string, string|int|bool> $recordData data to save, may be empty,
     *        but must not contain the keys "uid" or "usergroup"
     *
     * @return positive-int the UID of the new FE user
     */
    public function createAndLoginFrontEndUser($frontEndUserGroups = '', array $recordData = []): int
    {
        $frontEndUserUid = $this->createFrontEndUser($frontEndUserGroups, $recordData);

        $this->loginFrontEndUser($frontEndUserUid);

        return $frontEndUserUid;
    }

    /**
     * Creates a BE user record.
     *
     * @param array<string, string|int|bool> $recordData data to save, may be empty, but must not contain the key "uid"
     *
     * @return positive-int the UID of the new BE user
     *
     * @deprecated #1537 will be removed in oelib 6.0
     */
    public function createBackEndUser(array $recordData = []): int
    {
        if (isset($recordData['uid'])) {
            throw new \InvalidArgumentException('The column "uid" must not be set in $recordData.', 1331489905);
        }

        return $this->createRecordWithoutTableNameChecks('be_users', $recordData);
    }

    /**
     * Changes an existing dummy record and stores the new data for this
     * record. Only fields that get new values in $recordData will be changed,
     * everything else will stay untouched.
     *
     * The array with the new recordData must contain at least one entry, but
     * must not contain a new UID for the record. If you need to change the UID,
     * you have to create a new record!
     *
     * @param non-empty-string $table the name of the table
     * @param positive-int $uid the UID of the record to change
     * @param non-empty-array<string, string|int|bool|float> $rawData the values to be changed as key-value pairs
     *
     * @throws \InvalidArgumentException
     * @throws \BadMethodCallException
     */
    public function changeRecord(string $table, int $uid, array $rawData): void
    {
        $this->initializeDatabase();
        $this->assertTableNameIsAllowed($table);
        // @deprecated #1532 will be removed in oelib 6.0
        $dummyColumnName = $this->getDummyColumnName($table);
        // @phpstan-ignore-next-line We're testing for a contract violation here.
        if ($uid === 0) {
            throw new \InvalidArgumentException('The parameter $uid must not be zero.', 1331490003);
        }
        // @phpstan-ignore-next-line We're testing for a contract violation here.
        if ($rawData === []) {
            throw new \InvalidArgumentException('The array with the new record data must not be empty.', 1331490008);
        }
        if (isset($rawData['uid'])) {
            throw new \InvalidArgumentException(
                'The parameter $recordData must not contain changes to the UID of a record.',
                1331490017
            );
        }
        // @deprecated #1532 will be removed in oelib 6.0
        if (isset($rawData[$dummyColumnName])) {
            throw new \InvalidArgumentException(
                'The parameter $recordData must not contain changes to the field "' . $dummyColumnName .
                '". It is impossible to convert a dummy record into a regular record.',
                1331490024
            );
        }

        $dataToSave = $this->normalizeDatabaseRow($rawData);
        $this->getConnectionForTable($table)->update($table, $dataToSave, ['uid' => $uid, $dummyColumnName => 1]);
    }

    /**
     * Creates a relation between two records on different tables (so called
     * m:n relation).
     *
     * @param non-empty-string $table name of the m:n table to which the record should be added
     * @param positive-int $uidLocal UID of the local table
     * @param positive-int $uidForeign UID of the foreign table
     *
     * @throws \InvalidArgumentException
     */
    public function createRelation(string $table, int $uidLocal, int $uidForeign): void
    {
        $this->initializeDatabase();
        if (!$this->isNoneSystemTableNameAllowed($table)) {
            $allowedTables = \implode(',', $this->ownAllowedTables);
            $errorMessage = \sprintf('The table "%1$s" is not allowed. Allowed tables: %2$s', $table, $allowedTables);
            throw new \InvalidArgumentException($errorMessage, 1331490358);
        }

        // @phpstan-ignore-next-line We're testing for a contract violation here.
        if ($uidLocal <= 0) {
            throw new \InvalidArgumentException('$uidLocal must be > 0, but is: ' . $uidLocal, 1331490370);
        }
        // @phpstan-ignore-next-line We're testing for a contract violation here.
        if ($uidForeign <= 0) {
            throw new \InvalidArgumentException('$uidForeign must be > 0, but is: ' . $uidForeign, 1331490378);
        }

        // @deprecated #1532 will be removed in oelib 6.0
        $this->markTableAsDirty($table);

        $recordData = [
            'uid_local' => $uidLocal,
            'uid_foreign' => $uidForeign,
            'sorting' => $this->getRelationSorting($table, $uidLocal),
            // @deprecated #1532 will be removed in oelib 6.0
            $this->getDummyColumnName($table) => 1,
        ];

        $this->getConnectionForTable($table)->insert($table, $recordData);
    }

    /**
     * Creates a relation between two records based on the rules defined in TCA
     * regarding the relation.
     *
     * @param non-empty-string $tableName name of the table from which a relation should be created
     * @param positive-int $uidLocal UID of the record in the local table
     * @param positive-int $uidForeign UID of the record in the foreign table
     * @param non-empty-string $columnName name of the column in which the relation counter should be updated
     *
     * @throws \InvalidArgumentException
     * @throws \BadMethodCallException
     */
    public function createRelationAndUpdateCounter(
        string $tableName,
        int $uidLocal,
        int $uidForeign,
        string $columnName
    ): void {
        $this->initializeDatabase();
        $this->assertTableNameIsAllowed($tableName);
        // @phpstan-ignore-next-line We're testing for a contract violation here.
        if ($uidLocal <= 0) {
            throw new \InvalidArgumentException(
                '$uidLocal must be > 0, but actually is "' . $uidLocal . '"',
                1331490425
            );
        }
        // @phpstan-ignore-next-line We're testing for a contract violation here.
        if ($uidForeign <= 0) {
            throw new \InvalidArgumentException(
                '$uidForeign must be  > 0, but actually is "' . $uidForeign . '"',
                1331490429
            );
        }

        $tca = $this->getTcaForTable($tableName);
        $relationConfiguration = $tca['columns'][$columnName];

        if (!isset($relationConfiguration['config']['MM']) || ($relationConfiguration['config']['MM'] === '')) {
            throw new \BadMethodCallException(
                'The column ' . $columnName . ' in the table ' . $tableName .
                ' is not configured to contain m:n relations using a m:n table.',
                1331490434
            );
        }

        if (isset($relationConfiguration['config']['MM_opposite_field'])) {
            // Switches the order of $uidForeign and $uidLocal as the relation
            // is the reverse part of a bidirectional relation.
            $this->createRelationAndUpdateCounter(
                $relationConfiguration['config']['foreign_table'],
                $uidForeign,
                $uidLocal,
                $relationConfiguration['config']['MM_opposite_field']
            );
        } else {
            $this->createRelation(
                $relationConfiguration['config']['MM'],
                $uidLocal,
                $uidForeign
            );
        }

        $this->increaseRelationCounter($tableName, $uidLocal, $columnName);
    }

    /**
     * Returns the TCA for a certain table.
     *
     * @param non-empty-string $tableName the table name to look up
     *
     * @return array<array<string, mixed>> associative array with the TCA description for this table
     */
    private function getTcaForTable(string $tableName): array
    {
        if (!isset($GLOBALS['TCA'][$tableName])) {
            throw new \BadMethodCallException('The table "' . $tableName . '" has no TCA.', 1569701919);
        }

        return $GLOBALS['TCA'][$tableName];
    }

    /**
     * Deletes all dummy records that have been added through this framework.
     * For this, all records with the "is_dummy_record" flag set to 1 will be
     * deleted from all tables that have been used within this instance of the
     * testing framework.
     *
     * If you set $performDeepCleanUp to TRUE, it will go through ALL tables to
     * which the current instance of the testing framework has access. Please
     * consider well, whether you want to do this as it's a huge performance
     * issue.
     *
     * @param bool $performDeepCleanUp whether a deep clean up should be performed, may be empty
     *
     * @deprecated #1532 will be removed in oelib 6.0
     */
    public function cleanUp(bool $performDeepCleanUp = false): void
    {
        $this->cleanUpDatabase($performDeepCleanUp);
        $this->cleanUpWithoutDatabase();
    }

    /**
     * @deprecated #1532 will be removed in oelib 6.0
     */
    private function cleanUpDatabase(bool $performDeepCleanUp = false): void
    {
        $this->initializeDatabase();
        $this->cleanUpTableSet(false, $performDeepCleanUp);
        $this->cleanUpTableSet(true, $performDeepCleanUp);
    }

    /**
     * Cleans up..
     */
    public function cleanUpWithoutDatabase(): void
    {
        $this->discardFakeFrontEnd();
        WritableEnvironment::restoreCurrentScript();
        GeneralUtility::flushInternalRuntimeCaches();
        if (\is_array($this->serverVariablesBackup)) {
            $GLOBALS['_SERVER'] = $this->serverVariablesBackup;
            $this->serverVariablesBackup = null;
        }

        foreach ($this->getHooks() as $hook) {
            if (method_exists($hook, 'cleanUp')) {
                $hook->cleanUp($this);
            }
        }

        if ((new Typo3Version())->getMajorVersion() <= 11) {
            RootlineUtility::purgeCaches();
        }
    }

    /**
     * Deletes a set of records that have been added through this framework for
     * a set of tables (either the test tables or the allowed system tables).
     * For this, all records with the "is_dummy_record" flag set to 1 will be
     * deleted from all tables that have been used within this instance of the
     * testing framework.
     *
     * If you set $performDeepCleanUp to TRUE, it will go through ALL tables to
     * which the current instance of the testing framework has access. Please
     * consider well, whether you want to do this as it's a huge performance
     * issue.
     *
     * @param bool $useSystemTables whether to clean up the system tables (TRUE) or the non-system test tables (FALSE)
     * @param bool $performDeepCleanUp whether a deep clean up should be performed, may be empty
     *
     * @deprecated #1532 will be removed in oelib 6.0
     */
    private function cleanUpTableSet(bool $useSystemTables, bool $performDeepCleanUp): void
    {
        if ($useSystemTables) {
            $tablesToCleanUp = $performDeepCleanUp ? self::ALLOWED_SYSTEM_TABLES : $this->dirtySystemTables;
        } else {
            $tablesToCleanUp = $performDeepCleanUp ? $this->ownAllowedTables : $this->dirtyTables;
        }

        foreach ($tablesToCleanUp as $currentTable) {
            $dummyColumnName = $this->getDummyColumnName($currentTable);
            if (!$this->tableHasColumn($currentTable, $dummyColumnName)) {
                continue;
            }

            // Runs a DELETE query for each allowed table. A "one-query-deletes-them-all" approach was tested,
            // but we didn't find a working solution for that.
            $this->getConnectionForTable($currentTable)->delete($currentTable, [$dummyColumnName => 1]);
            $this->resetAutoIncrementLazily($currentTable);
        }

        $this->dirtyTables = [];
    }

    /**
     * Checks whether a table has a column "uid".
     *
     * @param non-empty-string $table
     *
     * @deprecated will be removed in oelib 6.0
     */
    private function tableHasColumnUid(string $table): bool
    {
        return $this->tableHasColumn($table, 'uid');
    }

    /**
     * Checks whether a table has a column with a particular name.
     *
     * @param non-empty-string $table the name of the table to check
     * @param string $column the column name to check
     *
     * @return bool
     */
    private function tableHasColumn(string $table, string $column): bool
    {
        if ($column === '') {
            return false;
        }

        $this->retrieveColumnsForTable($table);

        return isset(self::$tableColumnCache[$table][$column]);
    }

    /**
     * Retrieves and caches the column data for the table $table.
     *
     * If the column data for that table already is cached, this function does
     * nothing.
     *
     * @param non-empty-string $table the name of the table for which the column names should be retrieved
     */
    private function retrieveColumnsForTable(string $table): void
    {
        if (isset(self::$tableColumnCache[$table])) {
            return;
        }

        $connection = $this->getConnectionForTable($table);
        $query = 'SHOW FULL COLUMNS FROM `' . $table . '`';
        $queryResult = $connection->executeQuery($query);
        $columns = [];
        if (\method_exists($queryResult, 'fetchAllAssociative')) {
            /** @var array<string, string> $fieldRow */
            foreach ($queryResult->fetchAllAssociative() as $fieldRow) {
                $field = $fieldRow['Field'];
                $columns[$field] = $fieldRow;
            }
        } else {
            /** @var array<string, string> $fieldRow */
            foreach ($queryResult->fetchAll() as $fieldRow) {
                $field = $fieldRow['Field'];
                $columns[$field] = $fieldRow;
            }
        }

        self::$tableColumnCache[$table] = $columns;
    }

    // Functions concerning a fake front end

    /**
     * @return non-empty-string
     */
    public function getFakeFrontEndDomain(): string
    {
        return self::FAKE_FRONTEND_DOMAIN_NAME;
    }

    /**
     * @return non-empty-string
     */
    public function getFakeSiteUrl(): string
    {
        return 'http://' . $this->getFakeFrontEndDomain() . '/';
    }

    /**
     * Fakes a TYPO3 front end, using $pageUid as front-end page ID if provided.
     *
     * If $pageUid is zero, the front end will have not page UID.
     *
     * This function creates `$GLOBALS['TSFE']`.
     *
     * @param positive-int $pageUid UID of a page record to use
     *
     * @return positive-int the UID of the used front-end page
     *
     * @throws \InvalidArgumentException if $pageUid is < 0
     */
    public function createFakeFrontEnd(int $pageUid): int
    {
        /** @phpstan-ignore-next-line We are explicitly checking for contract violations here */
        if ($pageUid <= 0) {
            throw new \InvalidArgumentException('$pageUid must be > 0.', 1331490786);
        }

        $this->suppressFrontEndCookies();
        $this->discardFakeFrontEnd();

        $this->setPageIndependentGlobalsForFakeFrontEnd();
        $this->setRequestUriForFakeFrontEnd($pageUid);

        $frontEndUser = GeneralUtility::makeInstance(FrontendUserAuthentication::class);
        $request = new ServerRequest();
        if ((new Typo3Version())->getMajorVersion() <= 10) {
            $frontEndUser->start();
            $frontEndUser->fetchGroupData();
        } else {
            $frontEndUser->start($request);
            $frontEndUser->fetchGroupData($request);
        }
        if ((new Typo3Version())->getMajorVersion() <= 11) {
            $frontEndUser->unpack_uc();
        }

        $this->createDummySite($pageUid);
        $allSites = GeneralUtility::makeInstance(SiteConfiguration::class)->getAllExistingSites(false);
        $site = $allSites[self::SITE_IDENTIFIER] ?? null;
        if (!$site instanceof Site) {
            throw new \RuntimeException('Dummy site not found.', 1635024025);
        }
        $language = $site->getLanguageById(0);
        $frontEnd = GeneralUtility::makeInstance(
            TypoScriptFrontendController::class,
            GeneralUtility::makeInstance(Context::class),
            $site,
            $language,
            new PageArguments($pageUid, '', []),
            $frontEndUser
        );
        $GLOBALS['TSFE'] = $frontEnd;

        $frontEnd->fe_user = $frontEndUser;
        if ((new Typo3Version())->getMajorVersion() <= 10) {
            $frontEnd->id = (string)$pageUid;
        } else {
            $frontEnd->id = $pageUid;
        }
        $frontEnd->determineId($request);
        // @deprecated #1528 will be removed in oelib 6.0
        $frontEnd->tmpl = GeneralUtility::makeInstance(TemplateService::class);
        $frontEnd->config = [
            'config' => ['MP_disableTypolinkClosestMPvalue' => true, 'typolinkLinkAccessRestrictedPages' => true],
        ];

        // @deprecated #1532 will be removed in oelib 6.0
        if (\in_array('sys_template', $this->dirtySystemTables, true)) {
            try {
                $rootLine = GeneralUtility::makeInstance(RootlineUtility::class, $pageUid)->get();
            } catch (PageNotFoundException $e) {
                $rootLine = [];
            }

            // @deprecated #1528 will be removed in oelib 6.0
            $frontEnd->tmpl->runThroughTemplates($rootLine);
            $frontEnd->tmpl->generateConfig();
            $frontEnd->tmpl->loaded = true;

            Locales::setSystemLocaleFromSiteLanguage($frontEnd->getLanguage());
        }

        $frontEnd->newCObj();
        /** @var ContentObjectRenderer $contentObject */
        $contentObject = $frontEnd->cObj;
        $contentObject->setLogger(new NullLogger());
        if ((new Typo3Version())->getMajorVersion() >= 11) {
            $contentObject->setRequest($request);
        }

        $this->hasFakeFrontEnd = true;
        $this->logoutFrontEndUser();

        return $pageUid;
    }

    /**
     * Discards all site configuration files, and creates a new configuration file for a dummy site.
     *
     * Starting with TYPO3 10, we will be able to use `SiteConfiguration::createNewBasicSite()` for this.
     */
    private function createDummySite(int $pageUid): void
    {
        $siteConfigurationDirectory = Environment::getConfigPath() . '/sites/';
        GeneralUtility::rmdir($siteConfigurationDirectory, true);
        $configurationDirectoryForTestingDummySite = $siteConfigurationDirectory . self::SITE_IDENTIFIER;
        GeneralUtility::mkdir_deep($configurationDirectoryForTestingDummySite);

        $url = $this->getFakeSiteUrl();
        $contents =
            "rootPageId: $pageUid
base: '$url'
baseVariants: {  }
languages:
  -
    title: 'Englisch'
    enabled: true
    languageId: 0
    base: '/'
    typo3Language: 'default'
    locale: 'en_US.UTF-8'
    iso-639-1: 'en'
    navigationTitle: 'Englisch'
    hreflang: 'en-US'
    direction: 'ltr'
    flag: 'us'
errorHandling: {  }
routes: {  }";

        $file = $configurationDirectoryForTestingDummySite . '/config.yaml';
        \file_put_contents($file, $contents);
        if (!\is_readable($file)) {
            throw new \RuntimeException('Site config file "' . $file . '" could not be created.', 1634918114);
        }
    }

    private function setPageIndependentGlobalsForFakeFrontEnd(): void
    {
        if (!\is_array($this->serverVariablesBackup)) {
            $this->serverVariablesBackup = $GLOBALS['_SERVER'] ?? null;
        }

        GeneralUtility::flushInternalRuntimeCaches();
        unset($GLOBALS['TYPO3_REQUEST']);

        $hostName = $this->getFakeFrontEndDomain();
        $documentRoot = '/var/www/html/public';
        $relativeScriptPath = '/index.php';
        $absoluteScriptPath = $documentRoot . '/index.php';
        $server = &$GLOBALS['_SERVER'];

        $server['DOCUMENT_ROOT'] = $documentRoot;
        $server['HOSTNAME'] = $hostName;
        $server['HTTP'] = 'off';
        $server['HTTP_ACCEPT_ENCODING'] = 'gzip, deflate, br';
        $server['HTTP_ACCEPT_LANGUAGE'] = 'de,en-US;q=0.7,en;q=0.3';
        $server['HTTP_HOST'] = $hostName;
        $server['HTTP_REFERER'] = $this->getFakeSiteUrl();
        $server['HTTP_USER_AGENT'] = 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:93.0) Gecko/20100101 Firefox/93.0';
        $server['PHP_SELF'] = '/index.php';
        $server['QUERY_STRING'] = '';
        $server['REMOTE_ADDR'] = '127.0.0.1';
        $server['REMOTE_HOST'] = '';
        $server['REQUEST_SCHEME'] = 'http';
        $server['SCRIPT_FILENAME'] = $absoluteScriptPath;
        $server['SCRIPT_NAME'] = $relativeScriptPath;
        $server['SERVER_ADDR'] = '127.0.0.1';
        $server['SERVER_NAME'] = $hostName;
        $server['SERVER_SOFTWARE'] = 'Apache/2.4.48 (Debian)';

        WritableEnvironment::setCurrentScript($absoluteScriptPath);
    }

    private function setRequestUriForFakeFrontEnd(int $pageUid): void
    {
        $slug = '/';
        if ($pageUid > 0) {
            $slug .= $pageUid;
        }

        $GLOBALS['_SERVER']['REQUEST_URI'] = $slug;
    }

    /**
     * Discards the fake front end.
     *
     * This function nulls out $GLOBALS['TSFE']. In addition, any logged-in front-end user will be logged out.
     *
     * The page record for the current front end will _not_ be deleted by this
     * function, though.
     *
     * If no fake front end has been created, this function does nothing.
     */
    private function discardFakeFrontEnd(): void
    {
        if (!$this->hasFakeFrontEnd()) {
            return;
        }

        $this->logoutFrontEndUser();

        $GLOBALS['TSFE'] = null;
        unset(
            $GLOBALS['TYPO3_REQUEST'],
            $GLOBALS['TYPO3_CONF_VARS']['FE']['dontSetCookie'],
            $GLOBALS['TYPO3_CONF_VARS']['SYS']['Objects'][FrontendUserAuthentication::class]
        );

        $this->hasFakeFrontEnd = false;
    }

    /**
     * Returns whether this testing framework instance has a fake front end.
     *
     * @return bool TRUE if this instance has a fake front end, FALSE
     *                 otherwise
     */
    private function hasFakeFrontEnd(): bool
    {
        return $this->hasFakeFrontEnd;
    }

    /**
     * Makes sure that no FE login cookies will be sent.
     */
    private function suppressFrontEndCookies(): void
    {
        // avoid cookies from the phpMyAdmin extension
        $GLOBALS['PHP_UNIT_TEST_RUNNING'] = true;

        $GLOBALS['_POST']['FE_SESSION_KEY'] = '';
        $GLOBALS['_GET']['FE_SESSION_KEY'] = '';

        $GLOBALS['TYPO3_CONF_VARS']['SYS']['Objects'][FrontendUserAuthentication::class]
            = ['className' => UserWithoutCookies::class];
    }

    // FE user activities

    /**
     * Fakes that a front-end user has logged in.
     *
     * If a front-end user currently is logged in, he/she will be logged out
     * first.
     *
     * Note: To set the logged-in users group data properly, the front-end user
     *       and his groups must actually exist in the database.
     *
     * @param positive-int $userId UID of the FE user, must not necessarily exist in the database
     *
     * @throws \InvalidArgumentException
     * @throws \BadMethodCallException if no front end has been created
     */
    private function loginFrontEndUser(int $userId): void
    {
        // @phpstan-ignore-next-line We're testing for a contract violation here.
        if ($userId <= 0) {
            throw new \InvalidArgumentException('The user ID must be > 0.', 1331490798);
        }
        if (!$this->hasFakeFrontEnd()) {
            throw new \BadMethodCallException(
                'Please create a front end before calling loginFrontEndUser.',
                1331490812
            );
        }

        if ($this->isLoggedIn()) {
            $this->logoutFrontEndUser();
        }

        $mapper = MapperRegistry::get(FrontEndUserMapper::class);
        // loads the model from database if it is a ghost
        $mapper->existsModel($userId);
        $dataToSet = $mapper->find($userId)->getData();
        $dataToSet['uid'] = $userId;
        if (isset($dataToSet['usergroup'])) {
            /** @var Collection<FrontEndUserGroup> $userGroups */
            $userGroups = $dataToSet['usergroup'];
            $dataToSet['usergroup'] = $userGroups->getUids();
        }

        $this->suppressFrontEndCookies();

        $frontEndUser = $this->getFrontEndController()->fe_user;
        if ($frontEndUser instanceof FrontendUserAuthentication) {
            $frontEndUser->createUserSession(['uid' => $userId, 'disableIPlock' => true]);
            $frontEndUser->user = $dataToSet;
            if ((new Typo3Version())->getMajorVersion() <= 10) {
                $frontEndUser->fetchGroupData();
            } else {
                $frontEndUser->fetchGroupData(new ServerRequest());
            }
        }

        GeneralUtility::makeInstance(Context::class)->setAspect('frontend.user', new UserAspect($frontEndUser));
    }

    /**
     * Logs out the current front-end user.
     *
     * If no front-end user is logged in, this function does nothing.
     *
     * @throws \BadMethodCallException if no front end has been created
     */
    public function logoutFrontEndUser(): void
    {
        if (!$this->hasFakeFrontEnd()) {
            throw new \BadMethodCallException(
                'Please create a front end before calling logoutFrontEndUser.',
                1331490825
            );
        }
        if (!$this->isLoggedIn()) {
            return;
        }

        $this->suppressFrontEndCookies();
        $frontEndUser = $this->getFrontEndController()->fe_user;
        if ($frontEndUser instanceof FrontendUserAuthentication) {
            $frontEndUser->logoff();
        }

        FrontEndLoginManager::getInstance()->logInUser();
        GeneralUtility::makeInstance(Context::class)->setAspect('frontend.user', new UserAspect());
    }

    /**
     * Checks whether a FE user is logged in.
     *
     * @return bool TRUE if a FE user is logged in, FALSE otherwise
     *
     * @throws \BadMethodCallException if no front end has been created
     *
     * @internal
     */
    public function isLoggedIn(): bool
    {
        if (!$this->hasFakeFrontEnd()) {
            throw new \BadMethodCallException('Please create a front end before calling isLoggedIn.', 1331490846);
        }

        return (bool)GeneralUtility::makeInstance(Context::class)->getPropertyFromAspect('frontend.user', 'isLoggedIn');
    }

    // ----------------------------------------------------------------------
    // Various helper functions
    // ----------------------------------------------------------------------

    /**
     * Returns a list of all table names that are available in the current
     * database.
     *
     * @return list<non-empty-string> table names
     */
    private function getAllTableNames(): array
    {
        $this->retrieveTableNames();

        return \array_keys(self::$tableNameCache);
    }

    /**
     * Retrieves the table names of the current DB and stores them in self::$tableNameCache.
     *
     * This function does nothing if the table names already have been retrieved.
     */
    private function retrieveTableNames(): void
    {
        if (self::$tableNameCache !== []) {
            return;
        }

        $connection = $this->getConnectionPool()->getConnectionByName('Default');
        $query = 'SHOW TABLE STATUS FROM `' . $connection->getDatabase() . '`';
        $queryResult = $connection->executeQuery($query);
        $tableNames = [];
        if (\method_exists($queryResult, 'fetchAllAssociative')) {
            /** @var array<string, string|int|null> $tableInformation */
            foreach ($queryResult->fetchAllAssociative() as $tableInformation) {
                /** @var non-empty-string $tableName */
                $tableName = $tableInformation['Name'];
                $tableNames[$tableName] = $tableInformation;
            }
        } else {
            /** @var array<string, string|int|null> $tableInformation */
            foreach ($queryResult->fetchAll() as $tableInformation) {
                /** @var non-empty-string $tableName */
                $tableName = $tableInformation['Name'];
                $tableNames[$tableName] = $tableInformation;
            }
        }

        self::$tableNameCache = $tableNames;
    }

    /**
     * Generates a list of allowed tables to which this instance of the testing
     * framework has access to create/remove test records.
     *
     * The generated list is based on the list of all tables that TYPO3 can
     * access (which will be all tables in this database), filtered by prefix of
     * the extension to test.
     *
     * The array with the allowed table names is written directly to
     * `$this->ownAllowedTables`.
     */
    private function createListOfOwnAllowedTables(): void
    {
        $this->ownAllowedTables = [];
        $allTables = $this->getAllTableNames();
        $length = \strlen($this->tablePrefix);

        foreach ($allTables as $currentTable) {
            if (substr_compare($this->tablePrefix, $currentTable, 0, $length) === 0) {
                $this->ownAllowedTables[] = $currentTable;
            }
        }
    }

    /**
     * Checks whether the given table name is in the list of allowed tables for
     * this instance of the testing framework.
     */
    private function isOwnTableNameAllowed(string $table): bool
    {
        return \in_array($table, $this->ownAllowedTables, true);
    }

    /**
     * Checks whether the given table name is in the list of allowed
     * system tables for this instance of the testing framework.
     */
    private function isSystemTableNameAllowed(string $table): bool
    {
        return \in_array($table, self::ALLOWED_SYSTEM_TABLES, true);
    }

    /**
     * Checks whether the given table name is in the list of allowed tables for this instance of the testing framework.
     *
     * @param string $table the name of the table to check
     */
    private function isNoneSystemTableNameAllowed(string $table): bool
    {
        return $this->isOwnTableNameAllowed($table);
    }

    /**
     * Checks whether the given table name is in the list of allowed tables or allowed system tables,
     * and throws an exception if it is not.
     *
     * @throws \InvalidArgumentException
     */
    private function assertTableNameIsAllowed(string $table): void
    {
        $isAllowed = $this->isNoneSystemTableNameAllowed($table) || $this->isSystemTableNameAllowed($table);
        if (!$isAllowed) {
            $allowedTables = \implode(',', \array_merge(self::ALLOWED_SYSTEM_TABLES, $this->ownAllowedTables));
            $errorMessage = \sprintf('The table "%1$s" is not allowed. Allowed tables: %2$s', $table, $allowedTables);
            throw new \InvalidArgumentException($errorMessage, 1569784847);
        }
    }

    /**
     * Returns the name of the column that marks a record as a dummy record.
     *
     * On most tables this is "is_dummy_record", but on system tables like
     * "pages" or "fe_users", the column is called "tx_oelib_dummy_record".
     *
     * On additional tables, the column is built using $this->tablePrefix as
     * prefix e.g. "tx_seminars_is_dummy_record" if $this->tablePrefix =
     * "tx_seminars".
     *
     * @param non-empty-string $table the table name to look up
     *
     * @return non-empty-string the name of the column that marks a record as dummy record
     *
     * @deprecated #1532 will be removed in oelib 6.0
     */
    public function getDummyColumnName(string $table): string
    {
        $this->initializeDatabase();

        return $this->isSystemTableNameAllowed($table) ? 'tx_oelib_is_dummy_record' : 'is_dummy_record';
    }

    /**
     * Counts the dummy records in the table given by the first parameter $table
     * that match a given WHERE clause.
     *
     * @param non-empty-string $table the name of the table to query
     * @param array<string, string|int|bool> $criteria key-value pairs to match
     *
     * @return int<0, max> the number of records that have been found, will be >= 0
     *
     * @throws \InvalidArgumentException
     *
     * @deprecated will be removed in oelib 6.0
     */
    public function count(string $table, array $criteria = []): int
    {
        $this->initializeDatabase();
        $this->assertTableNameIsAllowed($table);

        $allCriteria = $criteria;
        $dummyColumn = $this->getDummyColumnName($table);
        $allCriteria[$dummyColumn] = 1;
        $query = $this->getQueryBuilderForTable($table)->count('*')->from($table);
        $query->getRestrictions()->removeAll();
        foreach ($allCriteria as $identifier => $value) {
            $query->andWhere($query->expr()->eq($identifier, $query->createNamedParameter($value)));
        }
        if (\method_exists($query, 'executeQuery')) {
            $result = $query->executeQuery();
        } else {
            $result = $query->execute();
            if (!$result instanceof ResultStatement) {
                throw new \UnexpectedValueException('Expected ResultStatement, got int instead.', 1646321756);
            }
        }

        if (\method_exists($result, 'fetchOne')) {
            $count = (int)$result->fetchOne();
        } else {
            $count = (int)$result->fetchColumn();
        }

        return $count >= 0 ? $count : 0;
    }

    /**
     * Checks whether there is a dummy record in the table given by the first
     * parameter $table that has the given UID.
     *
     * @param non-empty-string $table the name of the table to query
     * @param positive-int $uid the UID of the record to look up
     *
     * @deprecated will be removed in oelib 6.0
     */
    public function existsRecordWithUid(string $table, int $uid): bool
    {
        // @phpstan-ignore-next-line We are explicitly testing for a contract violation here.
        if ($table === '') {
            throw new \InvalidArgumentException('$table must not be empty.', 1569785503);
        }
        // @phpstan-ignore-next-line We are explicitly testing for a contract violation here.
        if ($uid <= 0) {
            throw new \InvalidArgumentException('$uid must be > 0.', 1331490872);
        }

        $this->initializeDatabase();
        $this->assertTableNameIsAllowed($table);

        $dummyColumn = $this->getDummyColumnName($table);
        $queryResult = $this->getConnectionForTable($table)
            ->select(['*'], $table, ['uid' => $uid, $dummyColumn => 1]);
        if (\method_exists($queryResult, 'fetchAllAssociative')) {
            $data = $queryResult->fetchAllAssociative();
        } else {
            $data = $queryResult->fetchAll();
        }

        return $data !== [];
    }

    /**
     * Eagerly resets the auto increment value for a given table to the highest existing UID + 1.
     *
     * @param non-empty-string $table the name of the table on which we're going to reset the auto increment entry
     *
     * @throws \InvalidArgumentException
     *
     * @deprecated #1529 will be removed in oelib 6.0
     * @see resetAutoIncrementLazily
     */
    public function resetAutoIncrement(string $table): void
    {
        $this->initializeDatabase();
        $this->assertTableNameIsAllowed($table);

        // Checks whether the current table qualifies for this method. If there
        // is no column "uid" that has the "auto_increment" flag set, we should
        // not try to reset this inexistent auto increment index to avoid DB
        // errors.
        if (!$this->tableHasColumnUid($table)) {
            return;
        }

        $newAutoIncrementValue = $this->getMaximumUidFromTable($table) + 1;

        // Updates the auto increment index for this table. The index will be
        // set to one UID above the highest existing UID.
        $connection = $this->getConnectionPool()->getConnectionByName('Default');
        $query = 'ALTER TABLE `' . $table . '` AUTO_INCREMENT=' . $newAutoIncrementValue . ';';
        $connection->executeQuery($query);
    }

    /**
     * Resets the auto increment value for a given table to the highest existing
     * UID + 1 if the current auto increment value is higher than a certain
     * threshold over the current maximum UID.
     *
     * The threshold is 100 by default and can be set using
     * `setResetAutoIncrementThreshold`.
     *
     * @param non-empty-string $table the name of the table on which we're going to reset the auto increment entry
     *
     * @deprecated #1529 will be removed in oelib 6.0
     * @see resetAutoIncrement
     */
    private function resetAutoIncrementLazily(string $table): void
    {
        $this->initializeDatabase();
        $this->assertTableNameIsAllowed($table);

        // Checks whether the current table qualifies for this method. If there
        // is no column "uid" that has the "auto_increment" flag set, we should
        // not try to reset this inexistent auto increment index to avoid
        // database errors.
        if (!$this->tableHasColumnUid($table)) {
            return;
        }
        $currentAutoIncrement = $this->getAutoIncrement($table);
        if (!\is_int($currentAutoIncrement)) {
            return;
        }

        if ($currentAutoIncrement > ($this->getMaximumUidFromTable($table) + $this->resetAutoIncrementThreshold)) {
            $this->resetAutoIncrement($table);
        }
    }

    /**
     * Sets the threshold for resetAutoIncrementLazily.
     *
     * @param positive-int $threshold
     *
     * @throws \InvalidArgumentException
     *
     * @see resetAutoIncrementLazily
     *
     * @deprecated #1529 will be removed in oelib 6.0
     */
    public function setResetAutoIncrementThreshold(int $threshold): void
    {
        // @phpstan-ignore-next-line We are explicitly testing for a contract violation here.
        if ($threshold <= 0) {
            throw new \InvalidArgumentException('$threshold must be > 0.', 1331490913);
        }

        $this->resetAutoIncrementThreshold = $threshold;
    }

    /**
     * Reads the highest UID for a database table.
     *
     * This function may only be called after that the provided table name
     * has been checked to be non-empty, valid and pointing to an existing
     * database table that has the "uid" column.
     *
     * @param non-empty-string $table the name of an existing table that has the "uid" column
     *
     * @return int<0, max> the highest UID from this table, will be >= 0
     *
     * @deprecated will be removed in oelib 6.0
     */
    private function getMaximumUidFromTable(string $table): int
    {
        $connection = $this->getConnectionForTable($table);
        $query = 'SELECT MAX(uid) AS uid FROM `' . $table . '`';
        $queryResult = $connection->executeQuery($query);
        if (\method_exists($queryResult, 'fetchAllAssociative')) {
            $data = $queryResult->fetchAllAssociative();
        } else {
            $data = $queryResult->fetchAll();
        }

        $maximumUid = (int)($data['uid'] ?? 0);

        return $maximumUid >= 0 ? $maximumUid : 0;
    }

    /**
     * Reads the current auto increment value for a given table.
     *
     * This function is only valid for tables that actually have an auto
     * increment value.
     *
     * @param non-empty-string $table the name of the table for which the auto increment value should be retrieved
     *
     * @return positive-int|null the auto_increment value of table $table, will be > 0, or null if the table has none
     *
     * @deprecated #1529 will be removed in oelib 6.0
     */
    public function getAutoIncrement(string $table): ?int
    {
        $this->initializeDatabase();
        $this->assertTableNameIsAllowed($table);

        $connection = $this->getConnectionForTable($table);
        $query = 'SHOW TABLE STATUS WHERE Name = \'' . $table . '\';';
        $queryResult = $connection->executeQuery($query);
        if (\method_exists($queryResult, 'fetchAssociative')) {
            $row = $queryResult->fetchAssociative();
        } else {
            $row = $queryResult->fetch();
        }

        $autoIncrement = \is_array($row) && \is_numeric($row['Auto_increment']) ? (int)$row['Auto_increment'] : null;

        return (\is_int($autoIncrement) && $autoIncrement > 0) ? $autoIncrement : null;
    }

    /**
     * Puts one or multiple table names on the list of dirty tables (which
     * represents a list of tables that were used for testing and contain dummy
     * records and thus are called "dirty" until the next cleanup).
     *
     * @param non-empty-string $tableNames the table name or a comma-separated list of table names
     *        to put on the list of dirty tables
     *
     * @throws \InvalidArgumentException
     *
     * @deprecated #1532 will be removed in oelib 6.0
     */
    public function markTableAsDirty(string $tableNames): void
    {
        $this->initializeDatabase();
        $this->assertTableNameIsAllowed($tableNames);

        /** @var non-empty-string $currentTable */
        foreach (GeneralUtility::trimExplode(',', $tableNames, true) as $currentTable) {
            if ($this->isNoneSystemTableNameAllowed($currentTable)) {
                $this->dirtyTables[$currentTable] = $currentTable;
            } elseif ($this->isSystemTableNameAllowed($currentTable)) {
                $this->dirtySystemTables[$currentTable] = $currentTable;
            }
        }
    }

    /**
     * Returns the next sorting value of the relation table which should be used.
     *
     * Note: This function does not take already existing relations in the
     * database - which were created without using the testing framework - into
     * account. So you always should create new dummy records and create a
     * relation between these two dummy records, so you're sure there aren't
     * already relations for a local UID in the database.
     *
     * @see https://bugs.oliverklee.com/show_bug.cgi?id=1423
     *
     * @param non-empty-string $table the relation table
     * @param positive-int $uidLocal UID of the local table
     *
     * @return positive-int the next sorting value to use
     */
    private function getRelationSorting(string $table, int $uidLocal): int
    {
        if (!isset($this->relationSorting[$table][$uidLocal])) {
            $this->relationSorting[$table][$uidLocal] = 0;
        }

        $this->relationSorting[$table][$uidLocal]++;

        return $this->relationSorting[$table][$uidLocal];
    }

    /**
     * Updates an integer field of a database table by one. This is mainly needed
     * for counting up the relation counter when creating a database relation.
     *
     * The field to update must be of type int.
     *
     * @param non-empty-string $tableName name of the table
     * @param positive-int $uid the UID of the record to modify
     * @param non-empty-string $fieldName the field name of the field to modify
     *
     * @throws \InvalidArgumentException
     * @throws \BadMethodCallException
     */
    private function increaseRelationCounter(string $tableName, int $uid, string $fieldName): void
    {
        $this->assertTableNameIsAllowed($tableName);
        if (!$this->tableHasColumn($tableName, $fieldName)) {
            throw new \InvalidArgumentException(
                'The table ' . $tableName . ' has no column ' . $fieldName . '.',
                1331490986
            );
        }

        $connection = $this->getConnectionForTable($tableName);
        $query = 'UPDATE ' . $tableName . ' SET ' . $fieldName . '=' . $fieldName . '+1 WHERE uid=' . $uid;
        $queryResult = $connection->executeQuery($query);
        $numberOfAffectedRows = $queryResult->rowCount();
        if ($numberOfAffectedRows === 0) {
            throw new \BadMethodCallException(
                'The table ' . $tableName . ' does not contain a record with UID ' . $uid . '.',
                1331491003
            );
        }

        // @deprecated #1532 will be removed in oelib 6.0
        $this->markTableAsDirty($tableName);
    }

    /**
     * Gets all hooks for this class.
     *
     * @return list<object> the hook objects, will be empty if no hooks have been set
     */
    private function getHooks(): array
    {
        if (self::$hooksHaveBeenRetrieved) {
            return self::$hooks;
        }

        /** @var array<array-key, class-string> $hookClasses */
        $hookClasses = (array)($GLOBALS['TYPO3_CONF_VARS']['EXTCONF']['oelib']['testingFrameworkCleanUp'] ?? []);
        foreach ($hookClasses as $hookClass) {
            self::$hooks[] = GeneralUtility::makeInstance($hookClass);
        }

        self::$hooksHaveBeenRetrieved = true;

        return self::$hooks;
    }

    /**
     * Purges the cached hooks.
     */
    public function purgeHooks(): void
    {
        self::$hooks = [];
        self::$hooksHaveBeenRetrieved = false;
    }

    /**
     * Returns the current front-end instance.
     *
     * This method must only be called when there is a front-end instance.
     *
     * @return TypoScriptFrontendController
     */
    private function getFrontEndController(): TypoScriptFrontendController
    {
        return $GLOBALS['TSFE'];
    }
}
